﻿using System;
using System.Data;
using System.Configuration;
using System.Linq;
using System.Web;
using System.Web.Security;
using System.Web.UI;
using System.Web.UI.HtmlControls;
using System.Web.UI.WebControls;
using System.Web.UI.WebControls.WebParts;
using System.Xml.Linq;
using System.Collections.Generic;
using WarfarinOnWeb.Code.DataAccess;
using System.Web.Hosting;
using System.IO;


namespace WarfarinOnWeb.Code.BusinessLogic
{
    
    public class DataPointGenerator
    {
        private Model model;
        private List<ModelAttribute> attributeList;
        private int noLastReadings;
        DataTable attColumnDetails;
        //arffFilePath should be taken from webconfig
        private string arffFilePath = HostingEnvironment.ApplicationPhysicalPath + "App_Data\\Test";
        private string extension = "arff";
        private string logFile = HostingEnvironment.ApplicationPhysicalPath + "App_Data\\Log.txt";

        public DataPointGenerator(Model m, List<ModelAttribute> attributeList)
        {
            DataAccessHome dataAccess = new DataAccessHome();
            this.model = m;
            this.attributeList = attributeList;
            this.noLastReadings = GetNoLastReadings();
        }
        public DataPointGenerator()
        {
        }
        private int GetNoLastReadings()
        {
            //int readingCount = 0;
            int prevNo=0;
            foreach (ModelAttribute ma in attributeList)
            {
                if (ma.AttributeName.Substring(0, 3) == "pre")
                {
                    if (prevNo < int.Parse(ma.AttributeName.Substring(4, 1)))
                        prevNo = int.Parse(ma.AttributeName.Substring(4, 1));
                }
            }
            //there must be always one
            return prevNo++;
        }
        public string GenerateData()
        {
            //int result = 0;
            string arffData=PopulateFileDefinition(attributeList);          
            arffData=CreateArffData(arffData);
            //DateTime centuryBegin = new DateTime(2009, 1, 1);
            //DateTime currentDate = DateTime.Now;
            //long elapsedTicks = currentDate.Ticks - centuryBegin.Ticks;
            //arffFilePath = arffFilePath + elapsedTicks.ToString() + "." + extension;
            arffFilePath = arffFilePath + "." + extension;
            WriteToFile(arffData,arffFilePath);
            //GenerateModelFromArffFile();
            
            return arffFilePath;
            
        }

        private string CreateArffData(string arffData)
        {
            DataTable dt = null;
            Dictionary<int, int> inrPositions=null;
            int lastReadingToProcess=0;
            DataAccessHome dataAccess = new DataAccessHome();
            List<int> patientList = null;
            
            if (model.ModelType  == "single")
            {
                // if it is a single model we only have
                // one patient in the list
                patientList = new List<int>();
                patientList.Add(model.PatientID);
            }
            else
            {
               patientList = dataAccess.PatientList();
            }
            
            int highestAttInterval = FindHighestInterval();
            string dataPoint = "";
            //int startPosition = 0;
            bool IsDeltaIn = IsDeltaInFunctionList();
            //no of last readings to be examined for one data point
            //increases with delta function
            if (IsDeltaIn)
            {
                this.noLastReadings++;
            }
            
            
            foreach (int patient in patientList)
            {
                
                dt = CreateDaylyDosageTable(patient);
                inrPositions = FindInrPositions(dt);
                lastReadingToProcess = FindLastReadingToProcess(inrPositions, highestAttInterval);
                dataPoint = ""; 
                
                //check if patient has as many lastreadings as needed for this model
                if (lastReadingToProcess>0)
                {
                    foreach (KeyValuePair<int, int> position in inrPositions)
                    {
                        if (position.Value >= lastReadingToProcess)
                        {   
                            dataPoint = CreateDataPoint(dt, position.Value,this.attributeList,false);
                            // add datapoint to the file
                            arffData = arffData + System.Environment.NewLine + dataPoint;
                        }
                        else
                        {
                            break;
                        }
                    }
                }

            }
            return arffData;

        }
        
        public string CreateDataPoint(DataTable dt,int startPosition,List<ModelAttribute> attributeList, bool prediction)
        {
            string dataPoint = "";
            int oldStartPosition = startPosition;
            DataRow dr = null;
            ModelAttribute outputAtt = null;
            foreach (ModelAttribute ma in attributeList)
            {
                //reset start position to the startposition method paramater 
                startPosition = oldStartPosition;
                if (ma.AttributeName.Substring(0, 3) == "pre")
                {
                    int prevNo = int.Parse(ma.AttributeName.Substring(4, 1));
                    startPosition = FindPosition(dt, prevNo, oldStartPosition);
                }
                //if we are making prediction datapoint then anything that has to do with INR column
                //has to start from previous INR since we still don't have new INR record(we are just predicting it)
                if (prediction && ma.ColumnName == "INR")
                {
                    startPosition = FindPosition(dt, 1, oldStartPosition);
                }
                if (!ma.IsOutput)
                {
                    //you can have a total interval with delta it needs to be day
                    if (ma.Function.ToLower() == "delta")
                    {
                        //calldelta function
                        dataPoint = dataPoint + ComputeDeltaValue(dt, ma.ColumnName, ma.Interval, startPosition) + ",";
                    }
                    else if (ma.Function.ToLower() == "none")
                    {
                        //just add the column value to the data point
                        //call CalculateValue
                        dr = dt.Rows[startPosition];
                        string columnName = GetSQLColumnNameFromAttribute(ma.ColumnName);
                        string dataType = GetSQLColumnDataTypeFromColumnName(ma.ColumnName);
                        if (dataType.ToLower().Trim() == "datetime")
                        {
                            dataPoint = dataPoint + string.Format("{0:d}", dr[columnName]) + ",";
                        }
                        else
                        {
                            dataPoint = dataPoint + dr[columnName].ToString().Trim() + ",";
                        }


                    }

                    else
                    {
                        
                        dataPoint = dataPoint + ComputeValue(dt, ma.ColumnName, ma.Function, ma.Interval, startPosition) + ",";
                        //call CalculateValue
                    }
                }
                else
                {
                    outputAtt = ma;
                }

            }
            //add output
            if (prediction)
            {
                dataPoint = dataPoint + "?";
            }
            else
            {

                if (outputAtt.Function.ToLower() == "delta")
                {
                    //calldelta function
                    dataPoint = dataPoint + ComputeDeltaValue(dt, outputAtt.ColumnName, outputAtt.Interval, startPosition) + ",";
                }
                else if (outputAtt.Function.ToLower() == "none")
                {
                    //just add the column value to the data point

                    //find sql name for the att column name
                    dr = dt.Rows[startPosition];
                    string columnName = GetSQLColumnNameFromAttribute(outputAtt.ColumnName);
                    dataPoint = dataPoint + dr[columnName].ToString().Trim();
                }
                else
                {
                    dataPoint = dataPoint + ComputeValue(dt, outputAtt.ColumnName, outputAtt.Function, outputAtt.Interval, startPosition);
                    //call CalculateValue
                }
            }
            
            return dataPoint;

        }
        private int FindLastReadingToProcess(Dictionary<int, int> inrPositions,int highestInterval)
        {
            int result = 0;
            //start from the oldest reading(first in the database)
            //and check if there is enough days before it to satisfy longest interval
            // if not move to the one after that. Once you get to the point when you have enough days to process
            // the reading move to the position of the as many number of reading as you have to have in one datapoint
            for (int i = inrPositions.Count-1; i>0; i--)
            {
                //check if the position number is higher then the logest interval
                if (inrPositions[i] >= highestInterval +1)
                {
                    //check if we have as many readings to take off
                    if (i > noLastReadings)
                        result = inrPositions[i - noLastReadings];
                    break;
                }
            }
            return result;
        }
        private Dictionary<int, int> FindInrPositions(DataTable dt)
        {
            Dictionary<int, int> InrPosition = new Dictionary<int, int>();
            int i = 0;
            int count = 0;
            for (i = dt.Rows.Count - 1; i >= 0; i--)
            {
                DataRow dr = dt.Rows[i];
                if (dr["INR"] != DBNull.Value)
                {
                    count++;
                    InrPosition.Add(count, i);
                    
                }
            }
            return InrPosition;

        }
        private int FindPositionOfFirstINRReading(DataTable dt)
        {
            int i = 0;
            for (i = dt.Rows.Count - 1; i >= 0; i--)
            {
                DataRow dr = dt.Rows[i];
                if (dr["INR"] != DBNull.Value)
                {
                    
                    break; 
                }
            }
            return i;
        }
        private int FindPosition(DataTable dt, int previousReading, int startPosition)
        {
            int count = 0;
            int i=0;
            //find position of the previous reading
            for ( i = startPosition ; i >= 0; i--)
            {
                DataRow dr = dt.Rows[i];
                if (dr["INR"] != DBNull.Value )
                {
                    if (count < previousReading)
                        count++;
                    else
                        break;

                }
            }
                return i;
        }

       
        private int FindCountofINRReadings(DataTable dt)
        {
            int count = 0;
            for (int i = dt.Rows.Count - 1; i >= 0; i--)
            {
                DataRow dr = dt.Rows[i];
                if (dr["INR"] != DBNull.Value)
                {
                    count++;
                }
            }
            return count;

        }
        private string GetSQLColumnDataTypeFromColumnName(string attColumnName)
        {
            string dataType = "";
            foreach (DataRow dr in attColumnDetails.Rows)
            {
                if (dr[0].ToString().Trim() == attColumnName)
                {
                    dataType = dr["SQLDataType"].ToString().Trim();
                    break;
                }
            }
            return dataType;

        }
        private string GetSQLColumnNameFromAttribute(string attColumnName)
        {
            string columnName = "";
            foreach (DataRow dr in attColumnDetails.Rows)
            {
                if (dr[0].ToString().Trim() == attColumnName)
                {
                    columnName = dr["SQLName"].ToString().Trim();
                    break;
                }
            }
            return columnName;
        }
        private string ComputeValue(DataTable dt, string nColumnName, string nFunction, int nInterval, int nStartingPoint)
        {
            //start from the one day before the INR reading was done
            if (nColumnName.ToLower().Trim() == "dose")
                nStartingPoint--;
            string expression = nFunction + "(" + nColumnName.Trim() + ")";
            if (nInterval == 0)
                return string.Format("{0:0.0}", dt.Compute(expression, ""));
            else
            {

                DataRow drStart = dt.Rows[nStartingPoint];
                DataRow drEnd = dt.Rows[nStartingPoint - nInterval];

                string filter = "ReadingDate <= '" + string.Format("{0:d}", drStart["ReadingDate"]) + "' AND ReadingDate > '" + string.Format("{0:d}", drEnd["ReadingDate"]) + "'";
                //string filter = "ReadingDate > '" + string.Format("{0:d}", drStart["ReadingDate"]) + "'";
                return string.Format("{0:0.0}", dt.Compute(expression, filter));
            }

        }
        private string ComputeDeltaValue(DataTable dt, string nColumnName, int nInterval, int nStartingPoint)
        {
            float firstSum = 0;
            float secondSum = 0;
            //start from the one day before the INR reading was done
            if (nColumnName.ToLower().Trim() == "dose")
                nStartingPoint--;
            DataRow drStart = dt.Rows[nStartingPoint];
            DataRow drEnd = dt.Rows[nStartingPoint - nInterval];
            DataRow drEnd1 = null;
            string expression = "Sum" + "(" + nColumnName.Trim() + ")";
            string filter = "ReadingDate <= '" + string.Format("{0:d}", drStart["ReadingDate"]) + "' AND ReadingDate > '" + string.Format("{0:d}", drEnd["ReadingDate"]) + "'";
            //string filter = "ReadingDate > '" + string.Format("{0:d}", drStart["ReadingDate"]) + "'";
            firstSum = Convert.ToSingle(dt.Compute(expression, filter));
            for (int i = nStartingPoint - 1; i >= nInterval; i--)
            {
                drStart = dt.Rows[i];
                if (drStart["INR"] != DBNull.Value)
                {
                    drEnd1 = dt.Rows[i - nInterval];
                    break;
                }
            }
            if (drEnd1 == null)
            {
                return "0.0";
            }
            else
            {
                //refresh filter
                filter = "ReadingDate <= '" + string.Format("{0:d}", drStart["ReadingDate"]) + "' AND ReadingDate > '" + string.Format("{0:d}", drEnd1["ReadingDate"]) + "'";
                secondSum = Convert.ToSingle(dt.Compute(expression, filter).ToString());
                return string.Format("{0:0.0}", (firstSum - secondSum));
            }

        }
        private int FindHighestInterval()
        {
            int result = 0;
            foreach (ModelAttribute ma in attributeList)
            {
                {
                    if (ma.Interval > result)
                        result = ma.Interval;
                }

            }
            return result;
        }
        private bool IsDeltaInFunctionList()
        {
            bool result = false ;
            foreach (ModelAttribute ma in attributeList)
            {
                
                    if (ma.Function == "delta")
                    {
                        result = true;
                        break;
                    }
                

            }
            return result;

        }
        public string PopulateFileDefinition(List<ModelAttribute> attributeList)
        {
            string arffData = "";
            //FileStream aFile = new FileStream(HostingEnvironment.ApplicationPhysicalPath + "App_Data\\Test.arff", FileMode.Create);
            //tw = new StreamWriter(HostingEnvironment.ApplicationPhysicalPath + "App_Data\\Test.arff");
            //aFile.Write("@relation Warfarin".ToCharArray(),0,17);
            string outputAttribute = "";
            arffData = arffData + "@relation Warfarin" + System.Environment.NewLine ;
            
            DataAccessHome dataAccess = new DataAccessHome();
            attColumnDetails = dataAccess.GetAttributeColumns();
            foreach (ModelAttribute att in attributeList)
            {
                foreach (DataRow dr in attColumnDetails.Rows)
                {
                    if (dr[0].ToString().Trim() == att.ColumnName)
                    {
                        if (!att.IsOutput)
                        {
                            //aFile.Write("@attribute " + att.AttributeName + " " + dr["ArffType"].ToString());
                            arffData = arffData + "@attribute " + att.AttributeName + " " + dr["ArffType"].ToString().Trim() + System.Environment.NewLine;
                            break;
                        }
                        else
                        {
                            outputAttribute = "@attribute " + att.AttributeName + " " + dr["ArffType"].ToString().Trim() + System.Environment.NewLine;
                            break;
                        }
                    }
                }

            }
            //put tha output attribute(class) on the end
            arffData = arffData + outputAttribute;

            //get ready for data entries
           
            arffData = arffData + "@data" ;
            return arffData;
            
        }
        public void WriteToFile(string arffdata,string arffFilePath)
        {
            
            TextWriter tw;

            tw = new StreamWriter( arffFilePath);
            tw.Write(arffdata);
            tw.Close();
            tw.Dispose();
        }


        public DataTable CreateDaylyDosageTable(int patientID)
        {
            int count = 0;
            DataRow row;
            DataAccessHome dataAccess = new DataAccessHome();
            DataTable table = dataAccess.GetDoseTable(patientID);

            DataTable dtResult = GetTableDefinition();



            //for each row in Dosage table
            for (int i = 0; i < table.Rows.Count; i++)
            {
                DataRow dr = table.Rows[i];
                //if it is a first or last row 
                if (i == 0 || i == table.Rows.Count - 1)
                {
                    //if (dr["PDose1"] == DBNull.Value)
                    //    dr["PDose1"] = 1;
                    row = dtResult.NewRow();
                    row = PopulateDataRow(row, dr);
                    //row["ReadingDate"] = Convert.ToDateTime(dr["ReadingDate"]);
                    //row["INR"] = Convert.ToSingle(dr["INR"]);
                    //row["Nominal_INR"] = dr["Nominal_INR"].ToString();
                    //row["Dose"] = Convert.ToSingle(dr["PDose1"]);
                    dtResult.Rows.Add(row);
                    //if last row make sure you create days for Dose2,3...
                    if (i == table.Rows.Count - 1)
                    {
                        //if it is a last reading we don't need to know what is happening after it
                        //not before next reading comes in so do nothing  here!!!

                        //DateTime newReadingDate = Convert.ToDateTime(dr["ReadingDate"]).AddDays(1);
                        //for (int m = 2; m <= 7; m++)
                        //{
                        //    if (dr["PDose" + m] == DBNull.Value)
                        //        break;
                        //    else
                        //    {
                        //        row = dtResult.NewRow();
                        //        row["Gender"] = (dr["Gender"].ToString().Trim() == "") ? "MSG" : dr["Gender"].ToString();
                        //        if (!(Convert.IsDBNull(dr["Target_INR_High"]))) row["Target_INR_High"] = Convert.ToSingle(dr["Target_INR_High"]);
                        //        if (!(Convert.IsDBNull(dr["Target_INR_Low"]))) row["Target_INR_Low"] = Convert.ToSingle(dr["Target_INR_Low"]);
                        //        if (!(Convert.IsDBNull(dr["Date_of_Birth"]))) row["Date_of_Birth"] = Convert.ToDateTime(dr["Date_of_Birth"]);
                        //        row["ReadingDate"] = newReadingDate;
                        //        row["INR"] = DBNull.Value;
                        //        row["Nominal_INR"] = "";
                        //        row["Dose"] = Convert.ToSingle(dr["PDose" + m]);
                        //        dtResult.Rows.Add(row);
                        //        newReadingDate = newReadingDate.AddDays(1);
                        //    }
                        //}
                    }
                }
                //all other rows except last one
                else
                {
                    //reset count
                    count = 0;
                    //add the row for that date, there must be at least PDose1 to add.
                    DataRow dr1 = table.Rows[i];
                    row = dtResult.NewRow();
                    row = PopulateDataRow(row, dr1);
                    //row["ReadingDate"] = Convert.ToDateTime(dr1["ReadingDate"]);
                    //row["INR"] = Convert.ToSingle(dr1["INR"]);
                    //row["Nominal_INR"] = dr1["Nominal_INR"].ToString();
                    //row["Dose"] = Convert.ToSingle(dr1["PDose1"]);
                    dtResult.Rows.Add(row);
                    count++;

                    DataRow dr2 = table.Rows[i + 1];

                    if (Convert.ToDateTime(dr1["ReadingDate"]).AddDays(1) < Convert.ToDateTime(dr2["ReadingDate"]))
                    {
                        DateTime newReadingDate = Convert.ToDateTime(dr1["ReadingDate"]).AddDays(1);

                        for (int ii = 2; ii <= 7; ii++)
                        {
                            if (newReadingDate < Convert.ToDateTime(dr2["ReadingDate"]))
                            {
                                //populate table with the dosage from the Dose1....Dose7 if there are values for it and get the pattern
                                if (dr1["PDose" + ii] != DBNull.Value)
                                {
                                    count++;
                                    row = dtResult.NewRow();
                                    row["Gender"] = (dr1["Gender"].ToString().Trim() == "") ? "MSG" : dr1["Gender"].ToString();
                                    if (!(Convert.IsDBNull(dr1["Target_INR_High"]))) row["Target_INR_High"] = Convert.ToSingle(dr1["Target_INR_High"]);
                                    if (!(Convert.IsDBNull(dr1["Target_INR_Low"]))) row["Target_INR_Low"] = Convert.ToSingle(dr1["Target_INR_Low"]);
                                    if (!(Convert.IsDBNull(dr1["Date_of_Birth"]))) row["Date_of_Birth"] = Convert.ToDateTime(dr1["Date_of_Birth"]);
                                    row["ReadingDate"] = newReadingDate;
                                    row["INR"] = DBNull.Value;
                                    row["Nominal_INR"] = "";
                                    row["Dose"] = Convert.ToSingle(dr1["PDose" + ii]);
                                    dtResult.Rows.Add(row);
                                    newReadingDate = newReadingDate.AddDays(1);
                                }
                                else
                                {
                                    break;
                                }
                            }
                        }
                        while (newReadingDate < Convert.ToDateTime(dr2["ReadingDate"]))
                        {
                            //Not all of the Doses in a week are populated ,there is a pattern  so use it
                            for (int j = 1; j <= count; j++)
                            {
                                if (newReadingDate < Convert.ToDateTime(dr2["ReadingDate"]))
                                {
                                    row = dtResult.NewRow();
                                    row["Gender"] = (dr1["Gender"].ToString().Trim() == "") ? "MSG" : dr1["Gender"].ToString();
                                    if (!(Convert.IsDBNull(dr1["Target_INR_High"]))) row["Target_INR_High"] = Convert.ToSingle(dr1["Target_INR_High"]);
                                    if (!(Convert.IsDBNull(dr1["Target_INR_Low"]))) row["Target_INR_Low"] = Convert.ToSingle(dr1["Target_INR_Low"]);
                                    if (!(Convert.IsDBNull(dr1["Date_of_Birth"]))) row["Date_of_Birth"] = Convert.ToDateTime(dr1["Date_of_Birth"]);
                                    if (!(Convert.IsDBNull(dr1["ReadingDate"]))) row["ReadingDate"] = newReadingDate;
                                    row["INR"] = DBNull.Value;
                                    row["Nominal_INR"] = "";
                                    row["Dose"] = Convert.ToSingle(dr1["PDose" + j]);
                                    dtResult.Rows.Add(row);
                                    newReadingDate = newReadingDate.AddDays(1);

                                }
                                else
                                {
                                    break;
                                }
                            }
                        }





                    }



                }


            }
            return dtResult;

        }

        private DataTable GetTableDefinition()
        {
            DataTable dtResult = new DataTable();
            DataColumn myDataColumn;


            myDataColumn = new DataColumn();
            myDataColumn.DataType = Type.GetType("System.String");
            myDataColumn.ColumnName = "Gender";
            dtResult.Columns.Add(myDataColumn);

            myDataColumn = new DataColumn();
            myDataColumn.DataType = Type.GetType("System.DateTime");
            myDataColumn.ColumnName = "Date_of_Birth";
            myDataColumn.DefaultValue = "01/01/1900";
            dtResult.Columns.Add(myDataColumn);

            myDataColumn = new DataColumn();
            myDataColumn.DataType = Type.GetType("System.Single");
            myDataColumn.ColumnName = "Target_INR_Low";
            dtResult.Columns.Add(myDataColumn);

            myDataColumn = new DataColumn();
            myDataColumn.DataType = Type.GetType("System.Single");
            myDataColumn.ColumnName = "Target_INR_High";
            dtResult.Columns.Add(myDataColumn);

            myDataColumn = new DataColumn();
            myDataColumn.DataType = Type.GetType("System.DateTime");
            myDataColumn.ColumnName = "ReadingDate";
            dtResult.Columns.Add(myDataColumn);

            myDataColumn = new DataColumn();
            myDataColumn.DataType = Type.GetType("System.Single");
            myDataColumn.ColumnName = "INR";
            dtResult.Columns.Add(myDataColumn);

            myDataColumn = new DataColumn();
            myDataColumn.DataType = Type.GetType("System.String");
            myDataColumn.ColumnName = "Nominal_INR";
            dtResult.Columns.Add(myDataColumn);

            myDataColumn = new DataColumn();
            myDataColumn.DataType = Type.GetType("System.Single");
            myDataColumn.ColumnName = "Dose";
            dtResult.Columns.Add(myDataColumn);
            return dtResult;
        }


        private DataRow PopulateDataRow(DataRow drTo, DataRow drFrom)
        {
            drTo["Gender"] = (drFrom["Gender"].ToString().Trim() == "") ? "MSG" : drFrom["Gender"].ToString();

            if (!(Convert.IsDBNull(drFrom["Target_INR_High"]))) drTo["Target_INR_High"] = Convert.ToSingle(drFrom["Target_INR_High"]);
            if (!(Convert.IsDBNull(drFrom["Target_INR_Low"]))) drTo["Target_INR_Low"] = Convert.ToSingle(drFrom["Target_INR_Low"]);
            if (!(Convert.IsDBNull(drFrom["Date_of_Birth"]))) drTo["Date_of_Birth"] = Convert.ToDateTime(drFrom["Date_of_Birth"]);
            if (!(Convert.IsDBNull(drFrom["ReadingDate"]))) drTo["ReadingDate"] = Convert.ToDateTime(drFrom["ReadingDate"]);
            if (!(Convert.IsDBNull(drFrom["INR"]))) drTo["INR"] = Convert.ToSingle(drFrom["INR"]);
            drTo["Nominal_INR"] = drFrom["Nominal_INR"].ToString();
            drTo["Dose"] = Convert.ToSingle(drFrom["PDose1"]);
            return drTo;
        }


    }
}

    

