﻿using System;
using System.Data;
using System.Configuration;
using System.Linq;
using System.Web;
using System.Web.Security;
using System.Web.UI;
using System.Web.UI.HtmlControls;
using System.Web.UI.WebControls;
using System.Web.UI.WebControls.WebParts;
using System.Xml.Linq;
using System.Collections;
using System.Collections.Generic;
using WarfarinOnWeb.Code.DataAccess;
using WarfarinOnWeb.Code.BusinessLogic;
using weka.core;
using weka.classifiers;
using weka.experiment;
using weka.classifiers.trees;
using java.util;
using System.Web.Hosting;

namespace WarfarinOnWeb.Code.BusinessLogic
{
    public class PredictionBuilder
    {
        private Model model;
        private List<ModelAttribute> attributeList;
        private DataTable dt;
        private float[] pattern;
        private int intervalLen;
        private string arffFilePath = HostingEnvironment.ApplicationPhysicalPath + "App_Data\\TestPrediction";
        private string modelDirectoryPath = HostingEnvironment.ApplicationPhysicalPath + "App_Data\\Model\\";
        private string extension = "arff";
        //add patern and period
        public PredictionBuilder()
        {
            //DataOperation dOperation = new DataOperation();
            //this.attributeList = dOperation.GetAttributeList(modelID);
            //this.model = dOperation.GetModel(modelID);
            //this.dt = dOperation.CreateDaylyDosageTable(model.PatientID);
                

        }

        public string GetPrediction(int modelID, float[] pattern, int intervalLen)
        {
            string result = "";
            string columnDataType = "";
            DataAccessHome dataAccess = new DataAccessHome();
            DataPointGenerator dataGenerator = new DataPointGenerator();
            string datapoint="";
            this.pattern = pattern;
            this.intervalLen = intervalLen;


                this.model = dataAccess.GetModel(modelID);
                this.attributeList = dataAccess.GetAttributeList(modelID);
                this.dt = dataGenerator.CreateDaylyDosageTable(model.PatientID);
                DataTable dtCopy = dt.Copy();
                dtCopy = AddDosesFromPattern(dtCopy);
                datapoint = dataGenerator.PopulateFileDefinition(attributeList);
                datapoint = datapoint + System.Environment.NewLine + dataGenerator.CreateDataPoint(dtCopy, dtCopy.Rows.Count - 1, attributeList, true);
                //geneate unique number fot the file
                //DateTime centuryBegin = new DateTime(2010, 1, 1);
                //DateTime currentDate = DateTime.Now;
                //long elapsedTicks = currentDate.Ticks - centuryBegin.Ticks;
                //arffFilePath = arffFilePath + elapsedTicks.ToString() + "." + extension;
                arffFilePath = arffFilePath + "." + extension;
                dataGenerator.WriteToFile(datapoint, arffFilePath);
                WekaUtil weka = new WekaUtil();
                result = weka.Predict(arffFilePath, model.ModelFileName).ToString();
                result = IntrepretResult(result);
                
            
            return result;
        }

        
        private DataTable AddDosesFromPattern(DataTable dataTable)
        {
            int count = 1;
            DataOperation dOperation = new DataOperation();
            DataRow row;
            
            DataRow rowToCopy;
            DataRow drStart = dataTable.Rows[dataTable.Rows.Count - 1];
            DateTime startDate = Convert.ToDateTime(drStart["ReadingDate"]);
            DateTime endDate = startDate.AddDays(this.intervalLen);
            DateTime movingDate = Convert.ToDateTime(drStart["ReadingDate"]);
            //instead being 0 as initialy set up assign it a right value
            drStart["Dose"] = pattern[0];
            for (int dose=1;dose<pattern.Length; dose++)
            {
                if (pattern[dose] != 0.0)
                {
                    movingDate = movingDate.AddDays(1);
                    if (movingDate < endDate)
                    {
                        count++;
                        row = dataTable.NewRow();
                        rowToCopy = dataTable.Rows[dataTable.Rows.Count - 1];
                        row = PopulateDataRow(row, rowToCopy);
                        //make sure INR is set to null
                        row["INR"] = DBNull.Value;
                        row["ReadingDate"] = movingDate;
                        //and dose is the next value in the pattern
                        row["Dose"] = pattern[dose];
                        dataTable.Rows.Add(row);
                    }
                }
                else
                {
                    break;
                }
                
            }

            movingDate = movingDate.AddDays(1);
            while (movingDate < endDate)
            {
                for (int j = 0; j < count; j++)
                {
                   
                    if (movingDate < endDate)
                    {
                        row = dataTable.NewRow();
                        rowToCopy = dataTable.Rows[dataTable.Rows.Count - 1];
                        row = PopulateDataRow(row, rowToCopy);
                        //make sure INR is set to null
                        row["INR"] = DBNull.Value;
                        row["ReadingDate"] = movingDate;
                        //and dose is the next value in the pattern
                        row["Dose"] = pattern[j];
                        dataTable.Rows.Add(row);
                        movingDate = movingDate.AddDays(1);
                    }
                    else
                    {
                        break;
                    }
                }
            }

            DataRow dr = dataTable.Rows[dt.Rows.Count - 1];
            //make sure INR is not null since the datapoint creation wouldn't pick the added rows.
            dr["INR"] = 0;

            return dataTable;
        }

        private DataRow PopulateDataRow(DataRow drTo, DataRow drFrom)
        {
            drTo["Gender"] = (drFrom["Gender"].ToString().Trim() == "") ? "MSG" : drFrom["Gender"].ToString();

            if (!(Convert.IsDBNull(drFrom["Target_INR_High"]))) drTo["Target_INR_High"] = Convert.ToSingle(drFrom["Target_INR_High"]);
            if (!(Convert.IsDBNull(drFrom["Target_INR_Low"]))) drTo["Target_INR_Low"] = Convert.ToSingle(drFrom["Target_INR_Low"]);
            if (!(Convert.IsDBNull(drFrom["Date_of_Birth"]))) drTo["Date_of_Birth"] = Convert.ToDateTime(drFrom["Date_of_Birth"]);
            if (!(Convert.IsDBNull(drFrom["ReadingDate"]))) drTo["ReadingDate"] = Convert.ToDateTime(drFrom["ReadingDate"]);
            if (!(Convert.IsDBNull(drFrom["INR"]))) drTo["INR"] = Convert.ToSingle(drFrom["INR"]);
            drTo["Nominal_INR"] = drFrom["Nominal_INR"].ToString();
            drTo["Dose"] = Convert.ToSingle(drFrom["Dose"]);
            return drTo;
        }

        private string GetSQLColumnDataTypeFromColumnName(string attColumnName)
        {
            string dataType = "";
            DataAccess.DataAccessHome dataAccess = new WarfarinOnWeb.Code.DataAccess.DataAccessHome();
            DataTable attColumnDetails = dataAccess.GetAttributeColumns();
            foreach (DataRow dr in attColumnDetails.Rows)
            {
                if (dr[0].ToString().Trim() == attColumnName)
                {
                    dataType = dr["ArffType"].ToString().Trim();
                    break;
                }
            }
            return dataType;

        }

        private string IntrepretResult(string result)
        {
            string datatype = "";
            string newresult = "";
            foreach (ModelAttribute ma in attributeList)
            {
                if (ma.IsOutput)
                {
                    datatype = GetSQLColumnDataTypeFromColumnName(ma.ColumnName);
                }
            }
            if (datatype.Trim() == "real")
            {
                newresult = string.Format("{0:f}", Convert.ToSingle(result));
            }
            else
            {
                switch (result)
                {
                    case "0":
                        newresult = "LOW";
                        break;
                    case "1":
                        newresult = "IN";
                        break;
                    case "2":
                        newresult = "HIGH";
                        break;
                    default:
                        newresult = "Error";
                        break;

                }
            }
            return newresult;
        }

    }
}
